@@ -30,7 +30,7 @@ except ImportError:
 import copr.exceptions as copr_exceptions
 from copr.v3 import (
     Client, config_from_file, CoprException, CoprRequestException,
-    CoprNoConfigException, CoprConfigException, CoprNoResultException,
+    CoprConfigException, CoprNoResultException, CoprAuthException,
 )
 from copr.v3.pagination import next_page
 from copr_cli.helpers import cli_use_output_format
@@ -45,15 +45,7 @@ if six.PY2:
 else:
     from urllib.parse import urljoin, urlparse
 
-if sys.version_info < (2, 7):
-    class NullHandler(logging.Handler):
-        def emit(self, record):
-            pass
-else:
-    from logging import NullHandler
-
 log = logging.getLogger(__name__)
-log.addHandler(NullHandler())
 
 
 ON_OFF_MAP = {
@@ -72,7 +64,7 @@ BOOTSTRAP_MAP = {
 
 no_config_warning = """
 ================= WARNING: =======================
-File '{0}' is missing or incorrect.
+File '{0}' is incorrect.
 See documentation: man copr-cli.
 Any operation requiring credentials will fail!
 ==================================================
@@ -132,55 +124,75 @@ def buildopts_from_args(args, progress_callback):
     return buildopts
 
 
+def requires_api_auth(func):
+    """
+    For operations that need authentication with server, check early that we
+    have either the user/token pair in hand, or that we can use the GSSAPI
+    session for authentication.  IOW, this decorator contacts the server
+    only if the GSSAPI is ON and preferred -- and the session cookie is not
+    initialized.
+    """
+    def wrapper(self, args):
+        # This call implies that either login+token is specified, GSSAPI
+        # session file exists, or (as a fallback) that we are able to do
+        # GSSAPI handshake and initialize the GSSAPI session.
+        self.client.base_proxy.auth_username()
+        return func(self, args)
+
+    wrapper.__doc__ = func.__doc__
+    wrapper.__name__ = func.__name__
+    return wrapper
+
+
 class Commands(object):
     def __init__(self, config_path):
         self.config_path = config_path or '~/.config/copr'
 
         try:
             self.config = config_from_file(self.config_path)
-        except (CoprNoConfigException, CoprConfigException) as ex:
+        except CoprConfigException as ex:
             sys.stderr.write(no_config_warning.format(self.config_path, ex))
-            self.config = {"copr_url": "http://copr.fedoraproject.org", "no_config": True}
+            self.config = {"copr_url": "http://copr.fedoraproject.org"}
 
+        if self.config.get("gssapi") is None:
+            # Contrary to what is set in python-copr, we set GSSAPI on by
+            # default (unless user explicitly says otherwise).
+            self.config["gssapi"] = True
+        self.config["connection_attempts"] = 3
         self.client = Client(self.config)
 
-    def requires_api_auth(func):
-        """ Decorator that checks config presence
+    @property
+    def username(self):
         """
-
-        def wrapper(self, args):
-            if "no_config" in self.config:
-                sys.stderr.write("Error: Operation requires api authentication\n")
-                sys.exit(6)
-
-            return func(self, args)
-
-        wrapper.__doc__ = func.__doc__
-        wrapper.__name__ = func.__name__
-        return wrapper
-
-    def check_username_presence(func):
-        """ Decorator that checks if username was provided
+        Get the username from config, or obtain it via auth_check (transitively
+        via GSSAPI).
         """
+        if self.config.get("username"):
+            return self.config["username"]
 
-        def wrapper(self, args):
-            if "no_config" in self.config and args.username is None:
-                sys.stderr.write(
-                    "Error: Operation requires username\n"
-                    "Pass username to command or create `{0}`\n".format(self.config_path))
-                sys.exit(6)
+        if self.config.get("gssapi"):
+            try:
+                return self.client.base_proxy.auth_username()
+            except CoprAuthException:
+                log.error("Failed to determine Copr username from authentication.")
+                raise
+        raise CoprConfigException(
+            "This operation tries to detect your username, but it is not "
+            "possible to find it in configuration, and GSSAPI is disabled "
+        )
 
-            if args.username is None and self.config["username"] is None:
-                sys.stderr.write(
-                    "Error: Operation requires username\n"
-                    "Pass username to command or add it to `{0}`\n".format(self.config_path))
-                sys.exit(6)
-
-            return func(self, args)
-
-        wrapper.__doc__ = func.__doc__
-        wrapper.__name__ = func.__name__
-        return wrapper
+    @property
+    def ownername(self):
+        """
+        Determine the project ownername (== username) when not specified on
+        commandline.
+        """
+        try:
+            return self.username
+        except:
+            log.error("This operation needs a project ownername specified, "
+                      "fallback to your username failed.")
+            raise
 
     def parse_name(self, name):
         m = re.match(r"([^/]+)/(.*)", name)
@@ -188,7 +200,7 @@ class Commands(object):
             owner = m.group(1)
             name = m.group(2)
         else:
-            owner = self.config["username"]
+            owner = self.ownername
         return owner, name
 
     def build_url(self, build_id):
@@ -206,7 +218,7 @@ class Commands(object):
         """
         m = re.match(r"(([^/]+)/)?([^/]+)/(.*)", path)
         if m:
-            owner = m.group(2) or self.config["username"]
+            owner = m.group(2) or self.username
             return owner, m.group(3), m.group(4)
         raise CoprException("Unexpected chroot path format")
 
@@ -270,11 +282,11 @@ class Commands(object):
         except KeyboardInterrupt:
             pass
 
-    def action_whoami(self, args):
+    def action_whoami(self, _args):
         """
         Simply print out the current user as defined in copr config.
         """
-        print(self.config["username"])
+        print(self.username)
 
     def action_new_webhook_secret(self, args):
         """
@@ -302,7 +314,10 @@ class Commands(object):
 
         :param args: argparse arguments provided by the user
         """
-        self.client.build_proxy.auth_check()
+
+        # Before we start uploading potentially large source RPM file, make sure
+        # that the user actually has a valid credentials.
+        self.client.base_proxy.auth_check()
 
         builds = []
         for pkg in args.pkgs:
@@ -397,16 +412,6 @@ class Commands(object):
             data[arg] = getattr(args, arg)
         return self.process_build(args, self.client.build_proxy.create_from_custom, data)
 
-    @requires_api_auth
-    def action_build_distgit(self, args):
-        """
-        Method called when the 'buildfedpkg' action has been selected by the user.
-
-        :param args: argparse arguments provided by the user
-        """
-        data = {"clone_url": args.clone_url, "committish": args.branch}
-        return self.process_build(args, self.client.build_proxy.create_from_scm, data)
-
     def process_build(self, args, build_function, data, bar=None, progress_callback=None):
         username, project_dirname = self.parse_name(args.copr_repo)
         projectname = project_dirname.split(':')[0]
@@ -576,7 +581,6 @@ class Commands(object):
         print(MockProfile(build_config))
 
 
-    @check_username_presence
     def action_list(self, args):
         """ Method called when the 'list' action has been selected by the
         user.
@@ -584,7 +588,7 @@ class Commands(object):
         :param args: argparse arguments provided by the user
 
         """
-        username = args.username or self.config["username"]
+        username = args.username or self.ownername
         projects = self.client.project_proxy.get_list(username)
         if not projects:
             sys.stderr.write("No copr retrieved for user: '{0}'\n".format(username))
@@ -835,9 +839,13 @@ class Commands(object):
 
     def action_list_package_names(self, args):
         ownername, projectname = self.parse_name(args.copr)
-        packages = self.client.package_proxy.get_list(ownername=ownername, projectname=projectname)
-        for package in packages:
-            print(package.name)
+        pagination = {"limit": 1000}
+        packages = self.client.package_proxy.get_list(ownername=ownername, projectname=projectname,
+                                                      pagination=pagination)
+        while packages:
+            for package in packages:
+                print(package.name)
+            packages = next_page(packages)
 
     def action_get_package(self, args):
         ownername, projectname = self.parse_name(args.copr)
@@ -1340,11 +1348,6 @@ def setup_parser():
             help="Build packages from SRPM generated by custom script")
     parser_build_custom.set_defaults(func="action_build_custom")
 
-    # create the parser for the "buildfedpkg" command
-    parser_build_distgit = subparsers.add_parser("buildfedpkg", parents=[parser_distgit_args_parent, parser_build_parent],
-                                                  help="DEPRECATED. Use SCM source type instead.")
-    parser_build_distgit.set_defaults(func="action_build_distgit")
-
     # create the parser for the "buildscm" command
     parser_build_scm = subparsers.add_parser("buildscm", parents=[parser_scm_args_parent, parser_build_parent],
                                               help="Builds package from Git/DistGit/SVN repository.")
@@ -1738,13 +1741,26 @@ def setup_parser():
     return parser
 
 
-def enable_debug():
-    logging.basicConfig(
-        level=logging.DEBUG,
-        format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
-        datefmt='%H:%M:%S'
-    )
-    sys.stderr.write("#  Debug log enabled  #\n")
+def setup_logging(debug):
+    """
+    Configure the global 'log' object so it prints to standard error output
+    (INFO+ messages).  When --debug is used, all logging (even external
+    libraries) is turned on for DEBUG+ level.
+    """
+    if debug:
+        logging.basicConfig(
+            level=logging.DEBUG,
+            format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
+            datefmt='%H:%M:%S'
+        )
+        sys.stderr.write("#  Debug log enabled  #\n")
+        return
+
+    stderr = logging.StreamHandler()
+    stderr.setLevel(logging.INFO)
+    formatter = logging.Formatter("%(levelname)s: %(message)s")
+    stderr.setFormatter(formatter)
+    log.addHandler(stderr)
 
 
 def str2bool(v):
@@ -1757,7 +1773,7 @@ def str2bool(v):
 
 
 def _handle_frontend_api_request_error(e, args):
-    if "Request is not in JSON format" not in str(e):
+    if "is not in JSON format" not in str(e):
         return
     if not args.debug:
         sys.stderr.write("Try 'copr-cli --debug' for more info.\n")
@@ -1778,13 +1794,13 @@ def _handle_frontend_api_request_error(e, args):
 
 
 def main(argv=sys.argv[1:]):
+    # pylint: disable=too-many-branches
     try:
         # Set up parser for global args
         parser = setup_parser()
         # Parse the commandline
         arg = parser.parse_args(argv)
-        if arg.debug:
-            enable_debug()
+        setup_logging(arg.debug)
 
         if not "func" in arg:
             parser.print_help()
@@ -1815,6 +1831,14 @@ def main(argv=sys.argv[1:]):
     except argparse.ArgumentTypeError as e:
         sys.stderr.write("\nError: {0}\n".format(e))
         sys.exit(2)
+    except CoprConfigException as err:
+        sys.stderr.write("\nError: Copr configuration error. {0}\n".format(err))
+        sys.exit(6)
+    except CoprAuthException as err:
+        sys.stderr.write("\nError: Operation requires API authentication. "
+                         "See the 'AUTHENTICATION' section in man copr-cli(1).\n"
+                         "\nError: {0}\n".format(err))
+        sys.exit(7)
     except CoprException as e:
         sys.stderr.write("\nError: {0}\n".format(e))
         sys.exit(3)
@@ -1822,10 +1846,6 @@ def main(argv=sys.argv[1:]):
         sys.stderr.write("\nError: {0}\n".format(e))
         sys.exit(5)
 
-        # except Exception as e:
-        # print "Error: {0}".format(e)
-        # sys.exit(100)
-
 
 if __name__ == "__main__":
     main()
